#define LASTCHANNEL 4
#define CH_1 0
#define CH_2 1
#define CH_3 2
#define CH_4 3
#define PWM2RAD 0.002
#define RISING_EDGE 1
#define FALLING_EDGE 0
#define MINONWIDTH 950
#define MAXONWIDTH 2075
#define MINOFFWIDTH 12000
#define MAXOFFWIDTH 24000

class  Receiver {
public:
  int receiverData[LASTCHANNEL];
  int transmitterCommand[LASTCHANNEL];
  int transmitterCommandSmooth[LASTCHANNEL];
  int transmitterZero[3];
  float transmitterSmooth[LASTCHANNEL];
  
  Receiver(void) {
    transmitterCommand[CH_1] = 1500;
    transmitterCommand[CH_2] = 1500;
    transmitterCommand[CH_3] = 1500;
    transmitterCommand[CH_4] = 1500;

    for (byte channel = CH_1; channel < LASTCHANNEL; channel++) {
      transmitterCommandSmooth[channel] = 1.0;
      transmitterSmooth[channel] = 1.0;
      transmitterZero[channel] = 1500;
    }
  }

  // ******************************************************************
  // The following function calls must be defined in any new subclasses
  // ******************************************************************
  virtual void initialize(void);
  virtual void read(void);

  
  // returns non-smoothed non-scaled ADC data in PWM full range 1000-2000 values
  const int getRaw(byte channel) {
    return receiverData[channel];
  }
  
  // returns raw but smoothed receiver(channel) in PWM
  const int getRawSmoothed(byte channel) {
    return transmitterCommandSmooth[channel];
  }
 
   // returns smoothed & scaled receiver(channel) in PWM values, zero centered
  const int getData(byte channel) {
    return transmitterCommand[channel];
  }
  
  // return the smoothed & scaled number of radians/sec in stick movement - zero centered
  const float getSIData(byte channel) {
    return ((transmitterCommand[channel] - transmitterZero[channel]) * (2.5 * PWM2RAD));  // +/- 2.5RPS 50% of full rate
  }

  // returns Zero value of channel in PWM
  const int getZero(byte channel) {
    return transmitterZero[channel];
  }

  // sets zero value of channel in PWM
  void setZero(byte channel, int value) {
    transmitterZero[channel] = value;
  }

  const float getSmoothFactor(byte channel) {
    return transmitterSmooth[channel];
  }

  void setSmoothFactor(byte channel, float value) {
    transmitterSmooth[channel] = value;
  }

  const float getAngle(byte channel) {
    // Scale 1000-2000 usecs to -45 to 45 degrees
    // m = 0.09, b = -135
    return (0.09 * transmitterCommand[channel]) - 135;
    //return (0.09 * receiverData[channel]) - 135;
  }
  float filterSmooth(float currentData, float previousData, float smoothFactor) {
    if (smoothFactor != 1.0) // only apply time compensated filter if smoothFactor is applied
      return (previousData * (1.0 - smoothFactor) + (currentData * smoothFactor)); 
    else
      return currentData; // if smoothFactor == 1.0, do not calculate, just bypass!
  }

};

/*************************************************/
/*************** AeroQuad PCINT ******************/
/*************************************************/
volatile uint8_t *port_to_pcmask[] = {
  &PCMSK0,
  &PCMSK1,
  &PCMSK2
};
volatile static uint8_t PCintLast[3];
// Channel data
typedef struct {
  byte edge;
  unsigned long riseTime;
  unsigned long fallTime;
  unsigned int  lastGoodWidth;
} tPinTimingData;
volatile static tPinTimingData pinData[8];

// Attaches PCINT to Arduino Pin
void attachPinChangeInterrupt(uint8_t pin) {
  uint8_t bit = digitalPinToBitMask(pin);
  uint8_t port = digitalPinToPort(pin);
  volatile uint8_t *pcmask;
  port -= 2;
  pcmask = port_to_pcmask[port];
  // set the mask
  *pcmask |= bit;
  // enable the interrupt
  PCICR |= 0x01 << port;
}

// ISR which records time of rising or falling edge of signal
static void measurePulseWidthISR(void) {
  uint8_t bit;
  uint8_t curr;
  uint8_t mask;
  uint8_t pin;
  uint32_t currentTime;
  uint32_t time;
#define port 2
  // get the pin states for the indicated port.
  curr = *portInputRegister(port+2);
  mask = curr ^ PCintLast[port];
  PCintLast[port] = curr;
  // mask is pins that have changed. screen out non pcint pins.
  if ((mask &= *port_to_pcmask[port]) == 0) {
    return;
  }
  currentTime = micros();
  // mask is pcint pins that have changed.
  for (uint8_t pin=4; pin < 8; pin++) {
    bit = 0x01 << pin;
    if (bit & mask) {
      // for each pin changed, record time of change
      if (bit & PCintLast[port]) {
        time = currentTime - pinData[pin].fallTime;
        pinData[pin].riseTime = currentTime;
        if ((time >= MINOFFWIDTH) && (time <= MAXOFFWIDTH))
          pinData[pin].edge = RISING_EDGE;
        else
          pinData[pin].edge = FALLING_EDGE; // invalid rising edge detected
      }
      else {
        time = currentTime - pinData[pin].riseTime;
        pinData[pin].fallTime = currentTime;
        if ((time >= MINONWIDTH) && (time <= MAXONWIDTH) && (pinData[pin].edge == RISING_EDGE)) {
          pinData[pin].lastGoodWidth = time;
          pinData[pin].edge = FALLING_EDGE;
        }
      }
    }
  }
#undef port
}


SIGNAL(PCINT2_vect) {
  measurePulseWidthISR(); // PORT D
}

// defines arduino pins used for receiver in arduino pin numbering schema
static byte receiverPin[4] = {4, 5, 6, 7}; // pins used 

class Receiver_4ch : public Receiver {
public:
  // Configure each receiver pin for PCINT
  void initialize() {
    for (byte channel = CH_1; channel < LASTCHANNEL; channel++) {
      pinMode(receiverPin[channel], INPUT);
      digitalWrite(receiverPin[channel], HIGH);
      pinData[receiverPin[channel]].edge = FALLING_EDGE;
      attachPinChangeInterrupt(receiverPin[channel]);
    }
  }

  // Calculate PWM pulse width of receiver data
  // If invalid PWM measured, use last known good time
  void read(void) {
    for(byte channel = CH_1; channel < LASTCHANNEL; channel++) {
      byte pin = receiverPin[channel];
      uint8_t oldSREG = SREG;
      cli();
      // Get receiver value read by pin change interrupt handler
      uint16_t lastGoodWidth = pinData[pin].lastGoodWidth;
      SREG = oldSREG;
      receiverData[channel] = lastGoodWidth;
      // Smooth the flight control transmitter inputs
      transmitterCommandSmooth[channel] = filterSmooth(receiverData[channel], transmitterCommandSmooth[channel], transmitterSmooth[channel]);
      //transmitterCommand[channel] = filterSmooth(receiverData[channel], transmitterCommandSmooth[channel], transmitterSmooth[channel]);
      transmitterCommand[channel] = transmitterCommandSmooth[channel];  
  }
 }
};





